/* Generic theme-independent CSS file */

terminal-window scrolledwindow undershoot.top,
terminal-window scrolledwindow undershoot.bottom,
terminal-window scrolledwindow overshoot.top,
terminal-window scrolledwindow overshoot.bottom
{
  background: none;
}

popover button.model
{
  padding-left: 16px;
  padding-right: 16px;
}

.disclosure-button
{
  padding-left: 4px;
  padding-right: 4px;
  min-width: 0;
}
